// Life.cpp, based on:
//         SaverModule.cp	Copyright 1996 Hiroshi Lockheimer
//         MrReflecto.cp	Copyright 1996 Andy Philpotts


# include	<CThread.h> 
# include	"Life.h"
# include	"CLifeAnimation.h"

# define	CELL_SIZE	(4)
# define	SNOOZE_TIME	(200000.0)

unsigned long kDefaultColorMap[COLOR_MAP_SIZE] = {
	0x00000000 /* Black */,
	0xDFDFDFDF /* Dark green */,
	0x2A2A2A2A /* Red */,
	0xFAFAFAFA /* Yellow */,
	0x20202020 /* Blue */
};	

const static BRect DUMMY_RECT( 0, 0, 5, 5 );
class CLifeView;


class CLifeThread : public CThread
{
private:
	CLifeView*		mLifeView;
	bool			mIsProcessorHog;
	
	virtual void	RunThread( );

public:
					CLifeThread(
						CLifeView*	inLifeView,
						bool		inIsProcessorHog );
						
	virtual void	Kill( );
	virtual void	Suspend( );
	void			SetProcessorHog( bool inIsProcessorHog );
};	


class CLifeView : public BView
{
public:
					CLifeView(
						BView*	inParentView,
						BRect	inBounds,
						bool	inIsProcessorHog );
	virtual			~CLifeView( );
	
	virtual void	AttachedToWindow( );
	virtual void	MessageReceived( BMessage* inMessage );
	virtual void	Show( );
	virtual void	Hide( );
	virtual void	Draw( BRect inUpdateRect );
	virtual	void	FrameResized( float inWidth, float inHeight );

	void			SetProcessorHog( bool inIsProcessorHog );
	void			Stop( );
		
private: friend class CLifeThread;
	void			WindowLock( );
	void			WindowUnlock( );

	BBitmap*		mBitmap;
	CLifeAnimation* mLifeAnimation;

	void			InitializeLifeBuffers( );
	void			FreeLifeBuffers( );

private:
	BWindow*		mParentWindow;
	CLifeThread		mLifeThread;
	bool			mIsDrawnAtLeastOnce;
};


CLifeView::CLifeView(
	BView*	inParentView,
	BRect	inBounds,
	bool	inIsProcessorHog )
:	mLifeThread( this, inIsProcessorHog ),
	BView( inBounds, "Life", B_FOLLOW_ALL, B_WILL_DRAW )
{
	mParentWindow = inParentView->Window( );
	mIsDrawnAtLeastOnce = FALSE;

	WindowLock( );
	inParentView->AddChild( this );
	InitializeLifeBuffers( );
	WindowUnlock( );
}


CLifeView::~CLifeView( )
{
	Stop( );
	FreeLifeBuffers( );
}


void CLifeView::AttachedToWindow( )
{
	BView::AttachedToWindow();
	SetViewColor(0,0,0);
	SetLowColor(0,0,0);
}


void CLifeView::MessageReceived( BMessage* inMessage )
{
	switch ( inMessage->what )
	{
		// XXX - Need preference changing code
		
		default:
			BView::MessageReceived( inMessage );
	}	
}


void CLifeView::Show( )
{
		// XXX - Do we want this routine?
	BView::Show( );
	mLifeThread.Resume( );
}


void CLifeView::Hide( )
{
		// XXX - Do we want this routine?
	mLifeThread.Suspend( );
	BView::Hide( );
}


void CLifeView::Draw( BRect inUpdateRect )
{
	BRect cellRect;

		// Start auto-drawing only after we've been drawn once
		// XXX - Is this what we want?
	if ( !mIsDrawnAtLeastOnce )
	{
		mIsDrawnAtLeastOnce = TRUE;
		mLifeThread.Resume( );
	}

	// Make sure Interface Kit has finished drawing any
	// bitmaps from the previous frame.
	Sync( );

	mLifeAnimation->RenderBitmap(
		(uchar*) mBitmap->Bits( ), mBitmap->BitsLength( ),
		mBitmap->BytesPerRow( ), CELL_SIZE / 4,
		kDefaultColorMap );
	
	DrawBitmapAsync( mBitmap );
	Flush( );
}


void CLifeView::FrameResized( float inWidth, float inHeight )
{
	FreeLifeBuffers( );
	InitializeLifeBuffers( );
}


void CLifeView::SetProcessorHog( bool inIsProcessorHog )
{
	mLifeThread.SetProcessorHog( inIsProcessorHog );
}


void CLifeView::Stop( )
{
		// XXX - The right thing to do? Lock and kill?
		// XXX - We start the thread so we can kill it successfully.
	WindowLock( );
	if ( !mIsDrawnAtLeastOnce )
		mLifeThread.Resume( );
	mLifeThread.Kill( );
	WindowUnlock( );
}


void CLifeView::WindowLock( )
{
	mParentWindow->Lock( );
}


void CLifeView::WindowUnlock( )
{
	mParentWindow->Unlock( );
}
	

void CLifeView::InitializeLifeBuffers( )
{
	WindowLock( );

	BRect bounds = Bounds( );
	bounds.OffsetTo( B_ORIGIN );

		// Calculate buffer dimensions	
		// XXX - What about fractional rounding?
	long pixelWidth		= bounds.right;
	long pixelHeight	= bounds.bottom;
	long bufferWidth	= pixelWidth / CELL_SIZE;
	long bufferHeight	= pixelHeight / CELL_SIZE;
	
		// Allocate buffers
	if ( !(mLifeAnimation = new CLifeAnimation( bufferWidth, bufferHeight )) )
		/* ERROR */;
	
	mBitmap = new BBitmap( BRect( 0, 0, pixelWidth, pixelHeight ), B_COLOR_8_BIT );
	if ( !mBitmap )
		/* ERROR */;
	
	WindowUnlock( );
}


void CLifeView::FreeLifeBuffers( )
{
	delete mLifeAnimation;
	if ( Window( ) )
		Sync( ); // before deleting bitmap!
	delete mBitmap;
}


CLifeThread::CLifeThread(
	CLifeView*	inLifeView,
	bool		inIsProcessorHog )
:	CThread( "Life", B_LOW_PRIORITY )
{
	mLifeView = inLifeView;
	mIsProcessorHog = inIsProcessorHog;
}


void CLifeThread::Kill( )
{
		// Get the window lock before we kill the thread.
		// This makes sure the window isn't left locked.
	mLifeView->WindowLock( );
	CThread::Kill( );
	mLifeView->WindowUnlock( );
}


void CLifeThread::Suspend( )
{
	mLifeView->WindowLock( );
	CThread::Suspend( );
	mLifeView->WindowUnlock( );
}


void CLifeThread::RunThread( )
{
	while ( 1 )
	{
		mLifeView->WindowLock( );
		mLifeView->mLifeAnimation->AdvanceOneFrame( );		
		mLifeView->Draw( mLifeView->Bounds( ) );
		mLifeView->WindowUnlock( );
	
		if ( !mIsProcessorHog )
			snooze( SNOOZE_TIME );
	}
}


void  CLifeThread::SetProcessorHog( bool inIsProcessorHog )
{
	mIsProcessorHog = inIsProcessorHog;
}


// ------------------------------------------------------------
// 		* module_initialize  [optional]
// ------------------------------------------------------------
// This function gets called once, immediately after saver has
// loaded your module
// It is not known at this point whether you are being loaded
// to save the screen or to configure yourself (or both)
// inSettings is a pointer to data inSettingsSize long that saver 
// has stored for you, it is your responsibility to free it before 
// your module is unloaded
// saver may pass you NULL and 0 if it cannot find any data that
// belongs to you

void module_initialize( void* inSettings, long inSettingsSize)
{

}


// ------------------------------------------------------------
// 		* module_cleanup  [optional]
// ------------------------------------------------------------
// This function gets called once, immediately before saver will
// unload your module
// This is the last chance to clean up after yourself
// Set outSettings to a pointer that is outSettingsSize long, 
// and saver will store it for you until the next time your module 
// is loaded
// (pass back NULL and 0 if you do not need to store anything)
// Do not attempt to free what you pass back to saver

void module_cleanup( void** outSettings, long* outSettingsSize)
{
	*outSettings = NULL;
	*outSettingsSize = 0;
}


// ------------------------------------------------------------
// 		* module_start_saving  [required]
// ------------------------------------------------------------
// This function gets called immediately after a BWindow
// covering the entire screen has been made visible
// The BWindow contains one BView (passed to you as inView)
// Spawn your thread(s) here and return from this function ASAP
// Do all your drawing in inView, or in its children (if you 
// decide to add any)
// inView's View and Low color are set to black, High color is white
// saver never locks inView's window for you, so be sure to
// inView->Window()->Lock() before you try to do anything with
// inView, otherwise you win a free trip to the debugger

CLifeView* lifeSaver;

void module_start_saving( BView	*inView )
{
	if ( !inView->Window( )->Lock( ) )
		return;
		
	lifeSaver = new CLifeView( inView, inView->Bounds( ), TRUE );
	if ( !lifeSaver )
		/* ERROR */;	

	inView->Window( )->Show( );
	inView->Window( )->Unlock( );
}


// ------------------------------------------------------------
// 		* module_stop_saving  [required]
// ------------------------------------------------------------
// This function gets called right before the BWindow that
// you have been drawing into gets closed
// Kill all your thread(s), free any memory that you no longer need, 
// unlock anything that you locked, be good and clean up after
// yourself

void module_stop_saving()
{
		// XXX - Remove our subview to keep Random happy. If
		// this isn't done, all hell breaks loose.
	BWindow* win = lifeSaver->Window( );
	if ( win->Lock( ) )
	{
		lifeSaver->Sync( );
		delete lifeSaver;
		win->Unlock( );
	}
}


// ------------------------------------------------------------
// 		* module_start_config  [required]
// ------------------------------------------------------------
// This function gets called when the user has selected your module
// in the settings panel
// You are free to add any children to inView or modify it
// in any way (but don't even think about deleting it)
// inView is 241 units wide and 244 units high

CLifeView* lifeConfigSample;
# define	CAPTION_HEIGHT	(15)

void module_start_config( BView* inView )
{
	if ( !inView->Window( )->Lock( ) )
		return;

	BRect bounds = inView->Bounds( );
	
	bounds.bottom -= CAPTION_HEIGHT;
	lifeConfigSample= new CLifeView( inView, bounds, FALSE );
	if ( !lifeConfigSample )
		/* ERROR */;
	
	bounds.top = bounds.bottom + 1;
	bounds.bottom = bounds.top + CAPTION_HEIGHT - 1;
	BStringView* caption = new BStringView( bounds, "Life Caption",
		"Life, by Eric Kidd" );
	if ( !caption )
		/* ERROR */;
	inView->AddChild( caption );
	caption->SetAlignment( B_ALIGN_CENTER );
	caption->SetViewColor( inView->ViewColor( ) );
	
	inView->Window( )->Unlock( );
}


// ------------------------------------------------------------
// 		* module_stop_config  [required]
// ------------------------------------------------------------
// This function gets called right before the BView that was passed
// to you in module_start_config gets deleted
// Poll the values of your controls, free any memory that you may
// have allocated, you know the game...
// Since your view will be deleted (mercilessly), it is not necessary
// to delete the children that you attached to it

void module_stop_config()
{
		// XXX - Stopping while attached to window to avoid race
		// condition of some sort
	lifeConfigSample->Stop( );
}

